local addonName, a = ...
local L = a.Localize
local s = SpellFlashAddon
local c = BittensGlobalTables.GetTable("BittensSpellFlashLibrary")
local bcm = a.BCM

local GetItemCount = GetItemCount
local IsMounted = IsMounted
local UnitBuff = UnitBuff

------------------------------------------------------------------------ Common
c.AddOptionalSpell("Mage Armor", nil, {
	Buff = "Mage Armor",
	BuffUnit = "player",
})

c.AddOptionalSpell("Conjure Mana Gem", nil, {
	NotIfActive = true,
	CheckFirst = function()
		if not s.Spec(1) and not s.HasTalent("Invocation") then
			return false
		end
		
		local count, target
		if c.HasGlyph("Mana Gem") then
			count = GetItemCount(c.GetID("Brilliant Mana Gem"), false, true)
			target = 7
		else
			count = GetItemCount(c.GetID("Mana Gem"), false, true)
			target = 3
		end 
		if c.IsSolo() then
			target = 1
		end
		return count < target
	end	
})

c.AddOptionalSpell("Mana Gem", nil, {
	Type = "item",
	CheckFirst = function()
		return a.ManaPercent < 80
			and not a.AlterTime
	end
})

c.AddOptionalSpell("Mana Gem", "at 10", {
	Type = "item",
	CheckFirst = function()
		return a.ManaPercent < 10
			and not a.AlterTime
	end
})

c.AddOptionalSpell("Brilliant Mana Gem", nil, {
	Type = "item",
	CheckFirst = function()
		return a.ManaPercent < 80
	end
})

c.AddOptionalSpell("Mirror Image")

c.AddOptionalSpell("Arcane Brilliance", nil, {
	NoRangeCheck = 1,
	CheckFirst = function()
		return c.RaidBuffNeeded(c.SPELL_POWER_BUFFS)
			or c.RaidBuffNeeded(c.CRIT_BUFFS)
	end
})

c.AddOptionalSpell("Dalaran Brilliance", nil, {
	NoRangeCheck = 1,
	CheckFirst = function()
		return c.RaidBuffNeeded(c.SPELL_POWER_BUFFS)
			or c.RaidBuffNeeded(c.CRIT_BUFFS)
	end
})

c.AddInterrupt("Counterspell")

c.AddOptionalSpell("Spellsteal", nil, {
	FlashColor = "aqua",
	CheckFirst = function()
		local unit = s.UnitSelection()
		if unit == nil or not s.Enemy(unit) then
			return false
		end
		
		for i = 1, 10000 do
			local _, _, _, _, _, _, _, _, isStealable, _, spellID
				= UnitBuff(unit, i)
			if spellID == nil then
				return false
			elseif isStealable then
				return true
			end
		end
	end
})

c.AddSpell("Nether Tempest", nil, {
	EarlyRefresh = 99,
	MyDebuff = "Nether Tempest",
	RunFirst = function(z)
		c.MakeOptional(z, c.IsAoE or c.IsSolo())
	end
})
c.ManageDotRefresh("Nether Tempest", 1)

c.AddSpell("Living Bomb", nil, {
	EarlyRefresh = 99,
	MyDebuff = "Living Bomb",
	RunFirst = function(z)
		c.MakeOptional(z, c.IsSolo())
	end
})
c.ManageDotRefresh("Living Bomb", 3)

c.AddSpell("Frost Bomb", nil, {
	NotIfActive = true,
	RunFirst = function(z)
		c.MakeOptional(z, c.IsSolo())
	end
})

c.AddOptionalSpell("Presence of Mind", nil, {
	NoGCD = true,
	CheckFirst = function()
		return not c.HasBuff("Alter Time", true)
	end
})

c.AddOptionalSpell("Evocation", nil, {
	CheckFirst = function()
		return a.ManaPercent <= 10 
			and not a.AlterTime
			and not c.HasBuff("Replenish Mana", false, false, "Mana Gem")
	end
})

c.AddOptionalSpell("Evocation", "for Health", {
	CheckFirst = function()
		return c.IsSolo()
			and c.HasGlyph("Evocation")
			and s.HealthPercent("player") < 60
	end
})

c.AddOptionalSpell("Evocation", "for Invoker's Energy", {
	CheckFirst = function()
		return c.HasTalent("Invocation")
			and c.GetBuffDuration("Invoker's Energy") < c.GetHastedTime(3)
	end
})

c.AddOptionalSpell("Evocation", "for Invoker's Energy at 5", {
	CheckFirst = function()
		return c.HasTalent("Invocation")
			and c.GetBuffDuration("Invoker's Energy", false, false, "Evocation") 
				< 5 + c.GetHastedTime(3)
			and not a.AlterTime
	end
})

c.AddOptionalSpell("Rune of Power", nil, {
	Buff = "Rune of Power",
	BuffUnit = "player",
	NoRangeCheck = true,
})

c.AddOptionalSpell("Rune of Power", "at 5", {
	NoRangeCheck = true,
	CheckFirst = function()
		return GetTime() - a.LastRuneCast > 5 + c.GetCastTime("Rune of Power") 
			and not a.AlterTime
	end
})

c.AddSpell("Fire Blast", nil, {
	NotWhileActive = true,
})

c.AddOptionalSpell("Cold Snap", nil, {
	CheckFirst = function()
		return c.GetHealthPercent("Player") < 70
	end
})

------------------------------------------------------------------------ Arcane
local function hasT6TalentFor(duration)
	return (not c.HasTalent("Rune of Power") or c.HasBuff("Rune of Power"))
		and (not c.HasTalent("Invocation")
			or c.GetBuffDuration("Invoker's Energy") >= duration)
		and (not c.HasTalent("Incanter's Ward") 
			or c.GetCooldown("Incanter's Ward") == 0
			or c.GetBuffDuration("Incanter's Absorption") > duration)
end

local function doPower()
	return a.MissilesStacks > 0
		and a.ChargeStacks > 2
		and not c.HasBuff("Alter Time")
		and hasT6TalentFor(15)
end

c.AddOptionalSpell("Arcane Power", nil, {
	CheckFirst = doPower,
})

c.AddOptionalSpell("Alter Time", "for Arcane", {
	CheckFirst = function(z)
		if c.HasBuff("Arcane Power") then
			z.FlashSize = nil
			return not c.HasBuff("Alter Time")
		elseif doPower() then
			z.FlashSize = s.FlashSizePercent() / 2
			return true
		end
	end
})

c.AddOptionalSpell("Evocation", "for Arcane", {
	CheckFirst = function()
		if c.HasBuff("Alter Time") then
			return false
		end
		
		return a.ManaPercent < 70
			or (c.HasTalent("Invocation") 
				and c.GetBuffDuration("Invoker's Energy") 
					< c.GetCastTime("Arcane Blast"))
	end
})

c.AddOptionalSpell("Evocation", "Interrupt", {
	FlashColor = "red",
	Override = function()
		return c.IsCasting("Evocation")
			and not c.HasTalent("Invocation") 
			and s.PowerPercent("player") > 98
	end
})

c.RegisterForFullChannels("Arcane Missiles", 2)

c.AddSpell("Arcane Missiles", nil, {
	CheckFirst = function(z)
		z.FlashColor = nil
		z.FlashSize = nil
		z.Continue = nil
		if c.HasBuff("Alter Time") or c.IsSolo() then
			return true
		end
		
		if a.MissilesStacks == 2 then
			if c.GetCooldown("Alter Time") == 0 then
				z.FlashSize = s.FlashSizePercent() / 2
				z.FlashColor = "yellow"
				z.Continue = true
			end
			return true
		end
		
		if a.ChargeStacks == 4 then
			if c.GetCooldown("Arcane Power") < 6 then
				z.FlashSize = s.FlashSizePercent() / 2
				z.FlashColor = "yellow"
				z.Continue = true
			end
			return true
		end
	end
})

c.AddOptionalSpell("Arcane Missiles", "for AoE")

c.AddSpell("Arcane Barrage", nil, {
	NoStopChannel = true,
	CheckFirst = function()
		return a.ChargeStacks == 4 and not c.HasBuff("Alter Time")
	end
})

c.AddSpell("Arcane Barrage", "for AoE", {
	NoStopChannel = true,
	CheckFirst = function()
		return a.ChargeStacks == 4
	end
})

c.AddSpell("Arcane Barrage", "if Losing Stacks", {
	FlashColor = "red",
	NoStopChannel = true,
	CheckFirst = function()
		local duration = a.ChargeDuration
		if c.HasBuff("Alter Time") then
			duration = math.min(duration, c.GetBuffDuration("Alter Time"))
		end
		return duration > 0 and duration < c.GetCastTime("Arcane Blast") + .1
	end
})

c.AddOptionalSpell("Flamestrike", nil, {
	NoRangeCheck = true,
	CheckFirst = function()
		return not c.IsCasting("Flamestrike")
	end
})

c.AddSpell("Arcane Explosion", nil, {
	Melee = true,
})

-------------------------------------------------------------------------- Fire
c.AssociateTravelTimes(.5, "Fireball", "Pyroblast", "Frostfire Bolt")
c.AssociateTravelTimes(.2, "Scorch", "Inferno Blast")

local function combustionCheckFirst(z, multiplier)
	local threshhold = c.GetOption("CombustionMin") * multiplier
	if bcm.PredictDamage(false, false) >= threshhold then
		if bcm.PredictDamage(true, false) >= threshhold then
			z.FlashColor = "yellow"
		else
			z.FlashColor = "red"
		end
		return true
	end
end

c.AddOptionalSpell("Molten Armor", nil, {
	Buff = "Molten Armor",
	BuffUnit = "player",
})

c.AddOptionalSpell("Combustion", nil, {
	NoGCD = true,
	CheckFirst = function(z)
		return combustionCheckFirst(z, 1)
	end
})

c.AddOptionalSpell("Combustion", "when Big", {
	NoGCD = true,
	CheckFirst = function(z)
		return combustionCheckFirst(z, 1.4)
	end
})

c.AddSpell("Pyroblast", nil, {
	CheckFirst = function()
		return a.PyroProc 
			or c.HasBuff("Presence of Mind") 
			or c.IsCasting("Presence of Mind")
	end
})

c.AddSpell("Inferno Blast", nil, {
	CheckFirst = function()
		return a.HeatingProc
	end
})

c.AddSpell("Scorch", nil, {
	Continue = true,
	CheckFirst = function()
		return s.Moving("player")
	end
})

c.AddSpell("Frostfire Bolt", nil, {
	Continue = true,
	CheckFirst = function()
		return c.HasGlyph("Frostfire Bolt")
	end
})

------------------------------------------------------------------------- Frost
c.AssociateTravelTimes(.7, "Frostbolt")

c.AddOptionalSpell("Summon Water Elemental", nil, {
    FlashColor = "yellow",
    CheckFirst = function()
        return not s.UpdatedVariables.PetAlive
            and not IsMounted()
    end
})

c.AddOptionalSpell("Frost Armor", nil, {
	Buff = "Frost Armor",
	BuffUnit = "player",
})

c.AddOptionalSpell("Evocation", "for Invoker's Energy for Frost", {
	CheckFirst = function()
		local dur = c.GetBuffDuration(
			"Invoker's Energy", false, false, "Evocation")
		return c.HasTalent("Invocation")
			and (dur == 0 
				or (a.VeinsCD < c.GetHastedTime(3) and dur < a.VeinsCD + 6))
	end
})

c.AddOptionalSpell("Rune of Power", "for Frost", {
	NoRangeCheck = true,
	CheckFirst = function()
		return not c.HasBuff("Rune of Power", false, false, true)
			or (a.VeinsCD < c.GetCastTime("Rune of Power")
				and a.LastRuneCast + 60 < a.VeinsCD + 26)
	end
})

c.AddOptionalSpell("Frozen Orb", nil, {
	NoRangeCheck = true,
	CheckFirst = function()
		return a.FingerCount == 0
	end
})

c.AddOptionalSpell("Icy Veins", nil, {
	CheckFirst = function(z)
		z.ID = a.VeinsID
		local atcd = c.GetCooldown("Alter Time")
		return (atcd < 4 or atcd > 49)
			and (a.FBStacks >= 3 or not c.HasGlyph("Icy Veins"))
	end
})

c.AddOptionalSpell("Presence of Mind", "for Frost", {
	NoGCD = true,
	CheckFirst = function(z)
		c.MakePredictor(
			z, c.HasGlyph("Icy Veins") and a.VeinsCD == 0, "yellow")
		local at = c.GetCooldown("Alter Time")
		return (at > 15 
				or (at == 0 
					and c.GetSpell("Alter Time for Frost"):CheckFirst())) 
			and (not c.HasGlyph("Icy Veins") 
				or a.VeinsCD > 15
				or (a.VeinsCD == 0 
					and c.GetSpell("Icy Veins"):CheckFirst()))
			and c.GetCastTime("Frostbolt") > 1
			and not c.HasBuff("Alter Time", true, false, true)
	end
})

c.AddOptionalSpell("Alter Time", "for Frost", {
	CheckFirst = function(z)
		local hasVeins = c.HasBuff("Icy Veins", false, false, true)
		c.MakePredictor(
			z, 
			(not hasVeins and a.VeinsCD == 0) 
				or (not c.HasBuff("Presence of Mind") 
					and c.GetCooldown("Presence of Mind") == 0), 
			"yellow")
		return (a.VeinsCD == 0 or hasVeins or a.VeinsCD > 45)
			and a.FBStacks >= 3 
			and (a.BrainFreeze or a.FingerCount > 0)
	end
})

c.AddOptionalSpell("Freeze", nil, {
    Type = "pet",
    NoRangeCheck = 1,
    NoGCD = true,
    CheckFirst = function()
        return a.FingerCount < 2 
        	and s.SpellCooldown(c.GetID("Freeze")) == 0
        	and not s.Boss() 
    end
})

c.AddOptionalSpell("Deep Freeze", nil, {
	CheckFirst = function()
		return a.FingerCount < 2 
			and not s.Boss() 
			and c.GetCooldown("Frozen Orb") < 50
	end
})

c.AddSpell("Frostfire Bolt", "Prime", {
	CheckFirst = function()
		return a.BrainFreeze 
			and (a.AlterTime or c.GetBuffDuration("Brain Freeze") < 2)
	end
})

c.AddSpell("Frostfire Bolt", "under Brain Freeze", {
	CheckFirst = function(z)
		c.MakeOptional(z, a.HoldProcs)
		c.MakeMini(z, a.HoldProcs)
		return a.BrainFreeze and not a.HoldForVeins
	end
})

c.AddSpell("Ice Lance", "when Frozen", {
	CheckFirst = function(z)
		c.MakeOptional(z, a.HoldProcs)
		c.MakeMini(z, a.HoldProcs)
		return a.FingerCount > 0
	end
})

c.AddSpell("Ice Lance", "Prime", {
	CheckFirst = function()
		return (a.FingerCount > 0 
				and (a.AlterTime or c.GetBuffDuration("Fingers of Frost") < 2))
			or (a.FingerCount < 2 
				and c.HasDebuff("Deep Freeze", false, false, true))
	end
})

c.AddSpell("Ice Lance", "before Cap", {
	CheckFirst = function(z)
		c.MakeOptional(z, a.HoldProcs)
		c.MakeMini(z, a.HoldProcs)
		return a.FingerCount == 2
			and (c.GetCooldown("Frozen Orb") > 50 
				or c.CountLandings("Frostbolt", -3, 10) > 0
				or c.CountLandings("Frostfire Bolt", -3, 10) > 0)
	end
})

c.AddSpell("Frostbolt", "for Debuff", {
	CheckFirst = function(z)
		return a.FBStacks < 3 and not c.IsSolo()
	end
})

--c.AddOptionalSpell("Frost Nova", nil, {
--	CheckFirst = function()
--		return c.IsSolo() 
--			and c.DistanceAtTheMost() <= 12 
--			and not c.HasDebuff("Deep Freeze")
--			and not s.Boss()
--	end
--})
